//
//
// hello.ftdi.44.echo.c
//
// 115200 baud FTDI character echo, with flash string
//
// set lfuse to 0x5E for 20 MHz xtal
//
// Neil Gershenfeld
// 12/8/10
//
// (c) Massachusetts Institute of Technology 2010
// This work may be reproduced, modified, distributed,
// performed, and displayed for any purpose. Copyright is
// retained and must be preserved. The work is provided
// as is; no warranty is provided, and users accept all
// liability.
//

#include <stdio.h>
#include <avr/io.h>
#include <util/delay.h>
#include <avr/pgmspace.h>

#define output(directions,pin) (directions |= pin) // set port direction for output
#define input(directions,pin) (directions &= (~pin)) // set port direction for output
#define set(port,pin) (port |= pin) // set port pin
#define clear(port,pin) (port &= (~pin)) // clear port pin
#define pin_test(pins,pin) (pins & pin) // test for port pin
#define bit_test(byte,bit) (byte & (1 << bit)) // test for bit set
#define bit_delay_time 8.5 // bit delay for 115200 with overhead
#define bit_delay() _delay_us(bit_delay_time) // RS232 bit delay
#define half_bit_delay() _delay_us(bit_delay_time/2) // RS232 half bit delay
#define char_delay() _delay_ms(10) // char delay

#define MORSE_UNIT 150
#define MORSE_DIT() _delay_ms(MORSE_UNIT)
#define MORSE_DAH() _delay_ms(MORSE_UNIT*3)
#define MORSE_GAP() _delay_ms(MORSE_UNIT)
#define MORSE_LETTERS_GAP() _delay_ms(MORSE_UNIT*3)
#define MORSE_WORDS_GAP() _delay_ms(MORSE_UNIT*7)

#define SERIAL_PIN_IN (1 << PA0)
#define SERIAL_PIN_OUT (1 << PA1)

#define LED_PIN (1 << PB2)

#define BUTTON_PIN (1 << PA7)

#define MAX_BUFFER 20

// read character into rxbyte on pins pin
//    assumes line driver (inverts bits)
void get_char(volatile unsigned char *pins, unsigned char pin, char *rxbyte) {
   *rxbyte = 0;
   while (pin_test(*pins,pin));

   // delay to middle of first data bit
   half_bit_delay();
   bit_delay();

   //unrolled loop, read data bits
   if pin_test(*pins,pin)
      *rxbyte |= (1 << 0);
   else
      *rxbyte |= (0 << 0);
   bit_delay();
   if pin_test(*pins,pin)
      *rxbyte |= (1 << 1);
   else
      *rxbyte |= (0 << 1);
   bit_delay();
   if pin_test(*pins,pin)
      *rxbyte |= (1 << 2);
   else
      *rxbyte |= (0 << 2);
   bit_delay();
   if pin_test(*pins,pin)
      *rxbyte |= (1 << 3);
   else
      *rxbyte |= (0 << 3);
   bit_delay();
   if pin_test(*pins,pin)
      *rxbyte |= (1 << 4);
   else
      *rxbyte |= (0 << 4);
   bit_delay();
   if pin_test(*pins,pin)
      *rxbyte |= (1 << 5);
   else
      *rxbyte |= (0 << 5);
   bit_delay();
   if pin_test(*pins,pin)
      *rxbyte |= (1 << 6);
   else
      *rxbyte |= (0 << 6);
   bit_delay();
   if pin_test(*pins,pin)
      *rxbyte |= (1 << 7);
   else
      *rxbyte |= (0 << 7);

   // wait for stop bit
   bit_delay();
   half_bit_delay();
}

// send character in txchar on port pin
//    assumes line driver (inverts bits)
void put_char(volatile unsigned char *port, unsigned char pin, char txchar) {

   // start bit
   clear(*port,pin);
   bit_delay();

   // unrolled loop to write data bits
   if bit_test(txchar,0)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,1)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,2)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,3)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,4)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,5)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,6)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,7)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();

   // stop bit
   set(*port,pin);
   bit_delay();

   // char delay
   bit_delay();
}

// print a null-terminated string
void put_string(volatile unsigned char *port, unsigned char pin, char *str) {
  static int index;
  index = 0;
  do {
    put_char(port, pin, str[index]);
    ++index;
  } while (str[index] != 0);
}


char* get_morse_code(char in_char) {
  switch (in_char) {
    case '0': return "_____";
    case '1': return ".____";
    case '2': return "..___";
    case '3': return "...__";
    case '4': return "...._";
    case '5': return ".....";
    case '6': return "_....";
    case '7': return "__...";
    case '8': return "___..";
    case '9': return "____.";
    case 'a': return "._";
    case 'b': return "_...";
    case 'c': return "_._.";
    case 'd': return "_..";
    case 'e': return ".";
    case 'f': return ".._.";
    case 'g': return "__.";
    case 'h': return "....";
    case 'i': return "..";
    case 'j': return ".___";
    case 'k': return "_._";
    case 'l': return "._..";
    case 'm': return "__";
    case 'n': return "_.";
    case 'o': return "___";
    case 'p': return ".__.";
    case 'q': return "__._";
    case 'r': return "._.";
    case 's': return "...";
    case 't': return "_";
    case 'u': return ".._";
    case 'v': return "..._";
    case 'z': return "__..";
    default: return "";
  }
}

void led_output_morse_code(char* morse_code) {
  for (uint8_t i = 0; morse_code[i] != 0; i++){
    set(PORTB, LED_PIN);
    switch (morse_code[i]) {
      case '.':
        MORSE_DIT();
        break;
      case '_':
        MORSE_DAH();
        break;
      default: //error
        MORSE_WORDS_GAP();
    }
    clear(PORTB, LED_PIN);
    MORSE_GAP();
  }
  MORSE_LETTERS_GAP();
}

void led_output_word_gap() {
  set(PORTB, LED_PIN);
  MORSE_WORDS_GAP();
  clear(PORTB, LED_PIN);
}

void led_output_morse_string(char* morse_string) {
  int index;
  index = 0;
  do {
    if ( morse_string[index] == ' '){
      led_output_word_gap();
    } else {
      char* morse_code = get_morse_code( morse_string[index]);
      led_output_morse_code(morse_code);
    }
    ++index;
  } while (morse_string[index] != 0);
  MORSE_WORDS_GAP();
}

void set_clock_divider_to_1() {
  CLKPR = (1 << CLKPCE);
  CLKPR = (0 << CLKPS3) | (0 << CLKPS2) | (0 << CLKPS1) | (0 << CLKPS0);
}

void initialize_pins(){
  output(DDRA, SERIAL_PIN_OUT);
  set(PORTA, SERIAL_PIN_OUT);

  input(DDRA, BUTTON_PIN);

  output(DDRB, LED_PIN);
  clear(PORTB, LED_PIN);
}

int main(void) {
  set_clock_divider_to_1();
  initialize_pins();

  put_string(&PORTA, SERIAL_PIN_OUT, "Morse LED\n");

  static char message_buffer[MAX_BUFFER] = {0};
  static char message_stored[MAX_BUFFER] = {0};
  static uint8_t buffer_index = 0;
  // static uint8_t times_long_button_pressed = 0;
  static char chr;

  for (;;) {
    get_char(&PINA, SERIAL_PIN_IN, &chr);
    if (chr) {
      if (++buffer_index < MAX_BUFFER-1) {
        message_buffer[buffer_index] = chr;
        message_buffer[buffer_index+1] = '\0';
        put_string(&PORTA, SERIAL_PIN_OUT, message_buffer);
        put_string(&PORTA, SERIAL_PIN_OUT, "\n");
      }
    }

    if (bit_is_clear(PINA, PINA7)) {
      set(PORTB, LED_PIN);
      uint8_t is_short_pression = 0;
      for (uint8_t i = 0; i < 40; i++) {
        _delay_ms(10);
        if (bit_is_set(PINA, PINA7)) {
          is_short_pression = 1;
          break;
        }
      }

      if (is_short_pression) {
        clear(PORTB, LED_PIN);
        put_string(&PORTA, SERIAL_PIN_OUT, "Output...");
        led_output_morse_string(message_stored);
        put_string(&PORTA, SERIAL_PIN_OUT, "done\n");
      } else {
        loop_until_bit_is_set(PINA, PINA7);
        clear(PORTB, LED_PIN);
        put_string(&PORTA, SERIAL_PIN_OUT, "Store buffer: ");
        memcpy(message_stored, message_buffer, MAX_BUFFER);
        put_string(&PORTA, SERIAL_PIN_OUT, message_stored);
        put_char(&PORTA, SERIAL_PIN_OUT, '\n');
      }
    }

  }
}
